#!/usr/bin/env python3
"""Stuff for doing stuff with Slack.

Slack: a FUSE implementation for Slack.

enumerate_tokens: a function that, given a login cookie, yields a list of Bearer tokens.
"""

__copyright__ = """
Copyright © 2020 Chris Danis

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import base64
import errno
import io
import logging
import os
import re
import stat
import time
import http.cookies
import urllib.parse
from collections import defaultdict

import cachetools
import refuse.high as fuse
import requests
from logzero import logger

import emojifs.utils


class Slack(fuse.LoggingMixIn, fuse.Operations):
    """A FUSE filesystem implementation for an individual Slack team."""
    def __init__(self, token: str, *, real_sizes: bool = True, name: str = ''):
        """Given an authentication token (xox.-....), construct a Slack instance."""

        # TODO: real_sizes=False is a big speedup on common tasks like `ls -l`, but, if you lie
        # about sizes you break reads and such, presumably unless you set direct_io (which disables
        # kernel buffer cache and probably introduces some other performance penalty for other
        # usages, although probably they're less bad/annoying).
        assert(real_sizes)

        self.name = name  # the identifying part before .slack.com
        self._token = token
        self._real_sizes = real_sizes
        self._base_url = f"https://{name}.slack.com/api/" if name else 'https://api.slack.com/api/'
        self._retry_after = {}  # URL -> time.time() after which it's ok to retry
        self._write_buffers = {}  # path (not name!) -> BytesIO
        self.__cached_metadata = cachetools.TTLCache(maxsize=1, ttl=600)  # for _get_all_emoji()

        self._unauthed_session = requests.Session()
        emojifs.utils.set_user_agent(self._unauthed_session.headers)

        self._session = requests.Session()
        emojifs.utils.set_user_agent(self._session.headers)
        self._session.headers['Authorization'] = f"Bearer {token}"

        # Our token should be usable for a single Slack.
        # To see if we have an admin token, first, we need our user ID.
        r = self._request('GET', self._url('auth.test'))
        j = r.json()
        self._user_id = j['user_id']
        self._base_url = j['url'] + 'api/'
        if not self.name:
            m = re.match(r'https://([^.]+)\.slack\.com/', j['url'])
            if m:
                self.name = m[1]
        r = self._request('GET', self._url('users.info'), params={'user': self._user_id})
        j = r.json()
        # TODO: we thought we cared about is_admin, but of course is_admin is sufficient but not
        # necessary to upload emoji.
        self._is_admin = (j['user']['is_admin'] or j['user']['is_owner']
                          or j['user']['is_primary_owner'])
        logger.info('👍 Successfully authenticated to %s as user %s', self.name, j['user']['name'])

    def _url(self, method):
        return f"{self._base_url}{method}"

    def _request(self, method, url, **kwargs):
        """Execute a request against our _session, respecting ratelimiting and raising if
        the response isn't okay.  Retry on ratelimiting (after sleeping) but not on other error."""
        # Respect any ratelimiting on the given URL path
        time.sleep(max(0, self._retry_after.get(url, 0) - time.time()))
        resp = self._session.request(method, url, **kwargs)
        if resp.status_code == 429:
            self._retry_after[url] = time.time() + resp.headers.get('retry-after', 60)
            logger.warn('Got ratelimited by Slack; retrying after %s seconds for %s',
                        self._retry_after[url], url)
            return self._request(method, url, **kwargs)
        else:
            resp.raise_for_status()
        logger.debug('resp for %s to %s json: %s', method, url, resp.json())
        assert(resp.json()['ok'])
        return resp

    def _request_all_pages(self, method, url, *, _paged_key: str, **kwargs):
        """Wrapper around _request for paginated APIs.  Will fetch all pages and return an array
        of results.  Must be provided _paged_key as a special arg telling the name of the key that
        contains the API results you want.  All other kwargs follow the usual Requests conventions
        (but it messes with params['page'] because it must).
        """
        accum = []
        pages_fetched = 0
        j = defaultdict(lambda: defaultdict(lambda: 1))  # Thanks to Guru Ibynaf.
        params = kwargs.pop('params', {})
        while pages_fetched < j['paging']['pages']:
            params['page'] = pages_fetched
            j = self._request(method, url, params=params, **kwargs).json()
            accum.extend(j[_paged_key])
            pages_fetched += 1
        return accum

    def _upload_emoji(self, name: str, file):
        """Upload an emoji to Slack.  file should be a file-like object (e.g. BytesIO)"""
        # The official API endpoint admin.emoji.add is *only for Enterprise*.  Sigh.
        try:
            self._request('POST', self._url('emoji.add'),
                          data={'mode': 'data', 'name': name},
                          files={'image': file})
        finally:
            self._invalidate_metadata()

    def _delete_emoji(self, name: str):
        """Delete the named emoji from Slack."""
        try:
            self._request('POST', self._url('emoji.remove'), data={'name': name})
        finally:
            self._invalidate_metadata()

    def _alias_emoji(self, src: str, dst: str):
        """Create an alias so that you can use :dst: as another name for :src:."""
        try:
            self._request('POST', self._url('emoji.add'),
                          data={'mode': 'alias', 'name': dst, 'alias_for': src})
        finally:
            self._invalidate_metadata()

    def _get_all_emoji(self):
        @cachetools.cached(self.__cached_metadata)
        def real(self):
            r = self._request_all_pages('GET', self._url('emoji.adminList'), _paged_key='emoji')
            return {e['name']: e for e in r}
        return real(self)

    def _invalidate_metadata(self):
        self.__cached_metadata.clear()

    def _path_to_name(self, path: str) -> str:
        """Translates a path-like string (e.g. umactually.png, parrotdad.gif) to an
        emoji name (umactually, parrotdad).  Path-like strings are allow to omit a suffix."""
        rv = path.split('/', maxsplit=1)[-1].split('.')[0]
        logger.debug("mapped '%s' to emoji name %s", path, rv)
        return rv

    def _emoji_to_filename(self, e, *, name: str = None) -> str:
        """Convert an emoji API dict to a pseudo-filename, including an extension."""
        real_name = e['name'] if not name else name
        if e['url'].startswith('http'):
            # grab the extension (gif/png) from the end of the URL
            rv = f"{real_name}.{e['url'].rsplit('.', maxsplit=1)[-1]}"
            logger.debug('mapped %s to file %s', real_name, rv)
            return rv
        elif e['url'].startswith('data:image/'):
            # grab the extension (gif/png) from the data URL MIME type
            extension = e['url'].split('data:image/')[1].split(';', maxsplit=1)[0]
            rv = f"{real_name}.{extension}"
            logger.debug('mapped %s to file %s (data:image/ URL)', e['url'], real_name, rv)
            return rv

    @cachetools.cached(cachetools.LRUCache(maxsize=100))
    def _get_emoji_bytes(self, url: str) -> bytes:
        """Returns the bytes for a given emoji URL.  Handles HTTP(S) and data URLs."""
        if url.startswith('http'):
            r = self._unauthed_session.get(url)
            r.raise_for_status()
            return r.content
        elif url.startswith('data:'):
            (prefix, data) = url.split(',', maxsplit=1)
            if not prefix.endswith('base64'):
                raise ValueError
            return base64.b64decode(data)

    @cachetools.cached(cachetools.LRUCache(maxsize=10000))
    def _get_content_length(self, url: str) -> int:
        """Returns the size of an emoji.  Handles HTTP(S) and data URLs."""
        if url.startswith('http'):
            r = self._unauthed_session.head(url)
            r.raise_for_status()
            # Slack emojis are served over CloudFront which provides Content-Length.
            return int(r.headers['Content-Length'])
        elif url.startswith('data:'):
            (prefix, data) = url.split(',', maxsplit=1)
            if not prefix.endswith('base64'):
                raise ValueError
            padding = data[-2:].count('=')
            return int(3*len(data)/4 - padding)

    # Now, the main course: FUSE operations implementations.

    def getattr(self, path, fh):
        emojis = self._get_all_emoji()

        # TODO: if we don't set allow_others in our fuse_main invocation,
        #       the 0o555 is probably just confusing.
        if path == '/':
            return dict(
                st_mode=stat.S_IFDIR | 0o555 | stat.S_IWUSR,
                st_mtime=max([e['created'] for e in emojis.values()]),
                st_ctime=min([e['created'] for e in emojis.values()]),
                st_atime=time.time(),
                st_nlink=2,
                st_uid=os.getuid(),
                st_gid=os.getgid(),
            )

        if path in self._write_buffers:
            return dict(
                st_mode=stat.S_IFREG | 0o600,
                st_atime=time.time(),
                st_ctime=time.time(),
                st_mtime=time.time(),
                st_nlink=1,
                st_uid=os.getuid(),
                st_gid=os.getgid(),
                st_size=len(self._write_buffers[path].getbuffer())
            )

        name = self._path_to_name(path)
        if name not in emojis:
            raise fuse.FuseOSError(errno.ENOENT)

        e = emojis[name]

        return dict(
            st_mode=(stat.S_IFLNK if e['is_alias'] else stat.S_IFREG) | 0o444,
            st_mtime=e['created'],
            st_ctime=e['created'],
            st_atime=time.time(),
            st_nlink=1,
            st_uid=os.getuid(),
            st_gid=os.getgid(),
            st_size=(self._get_content_length(e['url']) if self._real_sizes else 256*1024),
        )

    def readdir(self, path, fh=None):
        return ['.', '..'] + [self._emoji_to_filename(e) for e in self._get_all_emoji().values()]

    def readlink(self, path):
        emojis = self._get_all_emoji()
        name = self._path_to_name(path)
        if name not in emojis:
            raise fuse.FuseOSError(errno.ENOENT)
        e = emojis[name]
        if not e['is_alias']:
            raise fuse.FuseOSError(errno.EINVAL)
        # We don't 'deference' it ourselves using our metadata table, because it could be an alias
        # to a regular Unicode emoji.
        # This will represent it as a dangling symlink instead of throwing.
        # TODO: do something smart and/or reasonable about regular Unicode emoji
        #       (possibly, have another mountpoint for them?)
        return self._emoji_to_filename(e, name=e['alias_for'])

    def read(self, path, size, offset, fh):
        if path in self._write_buffers:
            b = self._write_buffers[path]
            b.seek(offset)
            return b.read(size)

        emojis = self._get_all_emoji()
        name = self._path_to_name(path)
        if name not in emojis:
            raise fuse.FuseOSError(errno.ENOENT)
        e = emojis[name]
        b = self._get_emoji_bytes(e['url'])
        return b[offset:offset+size]

    # TODO: override open() s.t. we don't allow non-create write modes.
    # TODO: eventually consider support re-writes of emoji, even though it's much more
    #       complicated on the client side.

    # Time for the scary stuff: Mutating operations!

    def unlink(self, path):
        emojis = self._get_all_emoji()
        name = self._path_to_name(path)
        if name not in emojis:
            raise fuse.FuseOSError(errno.ENOENT)
        self._delete_emoji(name)

    # TODO: both create() and symlink() need to check that the name being created is valid-ish
    #       (for now they'll just throw errors on release(), which may not reach the client)

    def create(self, path, mode, fi=None):
        self._write_buffers[path] = io.BytesIO()
        return 0

    def write(self, path, data, offset, fh):
        b = self._write_buffers[path]
        b.seek(offset)
        return self._write_buffers[path].write(data)

    # This is where the write() magic actually happens.  We need to make a single POST call, with
    # a valid and complete image file, so it's the only place where it really can.
    def release(self, path, fh):
        # release() is called for any opened file, including files that were only opened for read.
        if path in self._write_buffers:
            b = self._write_buffers[path]
            try:
                name = self._path_to_name(path)
                b.seek(0)
                self._upload_emoji(name, b)
            finally:
                b.close()
                del self._write_buffers[path]

    def truncate(self, path, length, fh=None):
        b = self._write_buffers[path]
        b.seek(0)
        b.truncate(length)

    def symlink(self, target_path, source_path):
        emojis = self._get_all_emoji()
        source = self._path_to_name(source_path)
        if source not in emojis:
            raise fuse.FuseOSError(errno.ENOENT)
        target = self._path_to_name(target_path)
        self._alias_emoji(source, target)


# Heavily inspired by https://github.com/emtunc/SlackPirate
# particularly the display_cookie_tokens function.
def enumerate_tokens(cookie: str):
    """Given a Slack login cookie string, enumerate all logged-in Slacks and
    return a list of tokens.

    cookie need only be the d= portion of the cookie."""

    rv = []

    try:
        cookies = http.cookies.SimpleCookie(cookie)
        if 'd' not in cookies:
            # Make a guess that we got passed only the value after `d=`.
            cookies = http.cookies.SimpleCookie()
            cookies['d'] = cookie
        cookies = dict(d=urllib.parse.quote(urllib.parse.unquote(cookies['d'].value)))
        sess = requests.Session()
        # a 'real' cookie jar was too annoying to figure out and seemed of dubious benefit anyway
        sess.headers['cookie'] = f"d={cookies['d']}"
        emojifs.utils.set_user_agent(sess.headers)
        # This Slack exists, but no one has access to it.
        # So, Slack helpfully lists all your logged-in teams.
        r = sess.get("https://emojifs-wasteland.slack.com")
        r.raise_for_status()
        ALREADY_SIGNED_IN_TEAM_REGEX = r"(https://[a-zA-Z0-9\-]+\.slack\.com)"
        SLACK_API_TOKEN_REGEX = r"\"api_token\":\"(xox[a-zA-Z]-[a-zA-Z0-9-]+)\""
        teams = (set(re.findall(ALREADY_SIGNED_IN_TEAM_REGEX, r.text))
                 - set(['https://status.slack.com', 'https://api.slack.com']))
        for team in teams:
            try:
                r = sess.get(team + "/customize/emoji")
                r.raise_for_status()
                parsed = re.findall(SLACK_API_TOKEN_REGEX, r.text)
                logger.debug('👀 Found %s tokens from %s', len(parsed), team)
                rv.extend(parsed)
            except Exception:
                logger.error("😖 Something went wrong when scraping token from %s", team, exc_info=1)
                # then continue
        return rv
    except Exception:
        logger.error("😖 Something went wrong when scraping login cookies", exc_info=True)
    finally:
        return rv


if __name__ == '__main__':
    # For debugging and/or manual testing in isolation.
    import argparse
    p = argparse.ArgumentParser()
    p.add_argument('mount')
    p.add_argument('token')
    args = p.parse_args()
    logging.basicConfig(level=logging.DEBUG)
    fuse = fuse.FUSE(Slack(token=args.token), args.mount, foreground=True)



#!/usr/bin/env python3

from emojifs import __version__, __repository__


def set_user_agent(x):
    """Sets a reasonable User-Agent. x should be a Request headers-like object."""
    x['User-Agent'] = f"emojifs/{__version__} ({__repository__}) {x['User-Agent']}"



#!/usr/bin/env python3
"""Stuff for gluing together multiple FUSE filesystems in a hierarchy.  Basically bind mounts."""

__copyright__ = """
Copyright © 2020 Chris Danis

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import bisect
import errno
import itertools
import os
import stat

import refuse.high as fuse
from logzero import logger


class Muxer(fuse.LoggingMixIn, fuse.Operations):
    def __init__(self, map):
        """Muxer is a FUSE filesystem compositor.  map should be a dict of path prefixes
        with values of FUSE implementations (e.g. Slack).  Muxer will dispatch operations
        to those filesystems based on path.

        The paths in map cannot be prefixes of one another.
        """
        def pairwise(iterable):
            "s -> (s0,s1), (s1,s2), (s2, s3), ..."
            a, b = itertools.tee(iterable)
            next(b, None)
            return zip(a, b)
        map = dict(sorted(map.items()))
        self._map = map
        self._mountpoints = list(map)
        # self._intermediates really should be a tree, but that's no fun.
        self._intermediates = set('/')
        for item in self._mountpoints:
            s = set((f"/{x}" for x in itertools.accumulate(filter(len, item.split('/')),
                                                           lambda *x: '/'.join(x))))
            self._intermediates = self._intermediates.union(s)
        self._intermediates = self._intermediates - set(self._mountpoints)
        assert not any(b.startswith(a) for (a, b) in pairwise(map))
        logger.info('🔧 Muxer created: map %s --> mountpoints %s intermediates %s', self._map,
                    self._mountpoints, self._intermediates)

    def _map_path(self, path):
        """Given a path, find the responsible FS in our map, and return a tuple of the path with
        its prefix stripped and the delegated FS."""
        def find_le(a, x):
            'Find rightmost value less than or equal to x'
            i = bisect.bisect_right(a, x)
            if i:
                return a[i-1]
            raise ValueError
        candidate = find_le(self._mountpoints, path)
        if not path.startswith(candidate):
            raise ValueError
        if path == candidate:
            return ('/', self._map[candidate])
        return (path[len(candidate):], self._map[candidate])

    # TODO: this should probably also passthrough init and destroy to all delegates

    # TODO: there must be a better way to do what follows...

    def getattr(self, path, *args, **kwargs):
        # Serve directory entries for our intermediates.
        if path in self._intermediates:
            return dict(
                st_mode=stat.S_IFDIR | 0o555,
                st_nlink=2,
                st_uid=os.getuid(),
                st_gid=os.getgid(),
            )
        # Delegate to mountpoints their / and below.
        try:
            (path, fs) = self._map_path(path)
        except ValueError:
            raise fuse.FuseOSError(errno.ENOENT)
        return getattr(fs, 'getattr')(path, *args, **kwargs)

    def readdir(self, path, *args, **kwargs):
        # For intermediate paths, synthesize all child intermediates and mountpoints.
        if path in self._intermediates:
            if not path.endswith('/'):
                path = path + '/'

            def find_children(ll):
                return [x[len(path):] for x in ll       # remove the parent's prefix from child
                        if x.startswith(path)           # by definition, child has parent's prefix
                        and '/' not in x[len(path):]    # maxdepth=1
                        and x != '/']                   # don't serve '/' ('.' is always served)

            return (['.', '..']
                    + find_children(self._intermediates)
                    + find_children(self._mountpoints))
        # Delegate to mountpoints their / and below.
        (path, fs) = self._map_path(path)
        return getattr(fs, 'readdir')(path, *args, **kwargs)

    def readlink(self, path, *args, **kwargs):
        (path, fs) = self._map_path(path)
        return getattr(fs, 'readlink')(path, *args, **kwargs)

    def read(self, path, *args, **kwargs):
        (path, fs) = self._map_path(path)
        return getattr(fs, 'read')(path, *args, **kwargs)

    def open(self, path, *args, **kwargs):
        (path, fs) = self._map_path(path)
        return getattr(fs, 'open')(path, *args, **kwargs)

    def unlink(self, path, *args, **kwargs):
        (path, fs) = self._map_path(path)
        return getattr(fs, 'unlink')(path, *args, **kwargs)

    def create(self, path, *args, **kwargs):
        (path, fs) = self._map_path(path)
        return getattr(fs, 'create')(path, *args, **kwargs)

    def write(self, path, *args, **kwargs):
        (path, fs) = self._map_path(path)
        return getattr(fs, 'write')(path, *args, **kwargs)

    def release(self, path, *args, **kwargs):
        (path, fs) = self._map_path(path)
        return getattr(fs, 'release')(path, *args, **kwargs)

    def truncate(self, path, *args, **kwargs):
        (path, fs) = self._map_path(path)
        return getattr(fs, 'truncate')(path, *args, **kwargs)

    def symlink(self, path, *args, **kwargs):
        (path, fs) = self._map_path(path)
        return getattr(fs, 'symlink')(path, *args, **kwargs)

    # TODO other methods?  e.g. xattrs eventually



#!/usr/bin/env python3
"""emojifs is a FUSE filesystem for manipulating Slack & Discord emoji.

⚠️  WARNING! ⚠️
😱☢️  DO NOT USE THIS PROGRAM. 😱☢️
This program is not a program of honor.

No highly esteemed function is executed here.

What is here is dangerous and repulsive to us.

The danger is still present, in your time, as it was in ours,
without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.

This program is best shunned and left unused (but it is free software,
and you are welcome to redistribute it under certain conditions).
😱☢️  DO NOT USE THIS PROGRAM. 😱☢️
"""

__copyright__ = """
Copyright © 2020 Chris Danis

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import functools
import logging
import operator
import os
import sys
from http.client import HTTPConnection

import logzero
import refuse.high as fuse
import tomlkit
from logzero import logger

import emojifs
import emojifs.slack
from emojifs.muxer import Muxer
from emojifs.slack import Slack


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    """Trivially combines RawDescriptionHelpFormatter with ArgumentDefaultsHelpFormatter."""
    pass


def _get(tree, branches, *, default=None):
    """Multi-level dict.get().  branches is an iterable of keys to traverse in tree."""
    try:
        return functools.reduce(operator.getitem, branches, tree)
    except KeyError:
        return default

def main():
    p = argparse.ArgumentParser(
        prog='emojifs',
        description=__doc__,
        formatter_class=CustomFormatter,
    )
    p.add_argument('-m', '--mountpoint', help='Where to mount emojifs.  If present here, '
                   'overrides mountpoint from config.')
    p.add_argument('-f', '--foreground', help='If set, stay in the foreground.', default=False)
    p.add_argument('-c', '--config', help='Path to your config file with secrets',
                   default=os.path.expanduser("~/.emojifs.toml"), type=argparse.FileType('r'))
    p.add_argument('-v', '--verbose', action='count', default=0,
                   help='Verbosity (-v, -vv, etc).  Higher verbosities will log all HTTP traffic '
                        '(NB: at higher levels, this will log your auth secrets!)')
    p.add_argument('-V', '--version', action='version',
                   version='%(prog)s {}'.format(emojifs.__version__))

    args = p.parse_args()

    loglevel = logging.WARNING
    if args.verbose >= 1:
        loglevel = logging.INFO
    if args.verbose >= 2:
        loglevel = logging.DEBUG
    if args.verbose >= 3:
        HTTPConnection.debuglevel = 1
        requests_log = logzero.setup_logger("urllib3", level=loglevel)
        requests_log.propagate = True

    logzero.setup_logger("fuse", level=loglevel)
    logzero.setup_logger("fuse.log-mixin", level=loglevel)
    logzero.loglevel(loglevel)

    # We parsed args, now parse the config file.

    config = tomlkit.parse(args.config.read())

    mountpoint = (args.mountpoint
                  or os.path.expanduser(_get(config, ['emojifs', 'mountpoint'], default='')))
    if not mountpoint:
        logger.error('A mountpoint must be specified either in the config file '
                     'or on the command line 😬')
        sys.exit(1)

    muxer_map = {}
    if 'slack' in config:
        slacks = {}
        slack_renames = _get(config, ['slack', 'renames'], default={})

        # TODO: maybe allow parsing a single string cookie from those sections? (or token?)

        def _add_slack_from_token(token: str):
            """Given a token, construct a Slack, fetch its associated name, then apply any
            name remappings, and stash it in our Slacks dict."""
            s = Slack(token=t)
            our_name = slack_renames.get(s.name, s.name)
            if our_name not in slacks:
                slacks[our_name] = s
                logger.debug("Added slack '%s' as '%s'", s.name, our_name)

        for t in _get(config, ['slack', 'tokens'], default=[]):
            _add_slack_from_token(t)

        for c in _get(config, ['slack', 'cookies'], default=[]):
            logger.info('🔑 trying login cookie scrape... 🥠')
            tokens = emojifs.slack.enumerate_tokens(c)
            for t in tokens:
                _add_slack_from_token(t)

        # assemble a slack_mounts dict to be passed to Muxer
        muxer_map = {f"/slack/{our_name}": s for (our_name, s) in slacks.items()}

    if 'discord' in config:
        logger.warning("Sorry, Discord is not yet supported 😞")
        # TODO discord
        pass

    mux = Muxer(muxer_map)

    foreground = args.foreground or _get(config, ['emojifs', 'foreground'], default=False)
    if args.verbose >= 1 and not foreground:
        logger.warn("You asked for verbose logging, but didn't also --foreground, "
                    "so I hope you only wanted logs during startup.  Seeya later 👋")
    logger.info('🚀 All systems go! About to mount at 🔜 %s (foreground %s)', mountpoint, foreground)
    try:
        fuse.FUSE(mux, mountpoint, foreground=foreground)
    except Exception:
        logger.error('😖 Something went wrong in FUSE setup', exc_info=True)


if __name__ == '__main__':
    main()
